load("tmp_input_heuristic_1.sage")

import subprocess
import multiprocessing
from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler

## To be replaced by the path to fplll
PATH_FPLLL = "/usr/bin/"

## Fixing the seed for reproductibility of results
set_random_seed(17);

## Define the number of threads for parallelization.
## For reproducibility of results, we fix nb_threads to 7
## For better performances, you may want to use all
## cores available on your machine (minus 1)
## (uncomment the second line for that)
nb_threads = 7
# nb_threads = multiprocessing.cpu_count()-1

precision = 2^53;
## Create the matrix of the lattice and reduce it 
## (to reduce the time complexity of the multiple cvp calls)
B_L = Matrix(ZZ,[[Integer(floor(precision*x)) for x in row] for row in list_L]);
B_L = B_L.BKZ(block_size = B_L.nrows())

## Wrapper for fplll to call with -a cvp
def __fplll_call(act, B, t=[], opts="", path = PATH_FPLLL):
    _t_fname = tmp_filename();
    _t_file  = open(_t_fname, 'w');
    _t_file.write("[{0}]\n".format("\n".join([str(list(i)).replace(",","") for i in B.rows()])));
    if (len(t) != 0):
        _t_file.write("{0}\n".format(str(list(t)).replace(",", "")));
    _t_file.close(); # Automatically erased when Sage exits

    # Requires Python >= 3.7
    #print ("cmd: env -i {0}./fplll -a {1} {2} {3}".format(path, act, opts, _t_fname));
    proc = subprocess.run(["env -i {0}./fplll -a {1} {2} {3}".format(path, act, opts, _t_fname)],
                          shell=True, stdout=subprocess.PIPE, text=True);
    # Valid if output is with commas (Sage compliant), otherwise matrices will be torn apart.
    out  = proc.stdout.rstrip().replace(",\n", ",").split("\n");
    
    assert (out[0] != '[]'); # Empty output can hide a precision problem.
    return out;


def call_fplll(B, t, path=PATH_FPLLL):
    # Weird bug? Output '[]' with -of c, and some vector without it. Doc says "-of c (default if -a cvp)".
    [_v_str] = __fplll_call("cvp", B, t=t, opts="", path = path);
    _v       = sage_eval("vector({0})".format(_v_str.replace(" ", ",")));
    return _v;
    

## Compute an approximation of the covering radius, by taking nb_tests random targets
## sampled using distribution D (everithing is multiplied by the
## precision to work with integers) and computing a closest vector.
## Define the approximate covering radius as the max of the distance obtained
## and the covering radius in infinity norm as the max of the infinity norm obtained
## (between the target and the closest vector in l2 norm)

@parallel(nb_threads)
def approx_covering_radius(B_L,nb_tests, D, seed_randomness):
   set_random_seed(seed_randomness); ## change the randomness for the different processes
   mu_2 = 0
   mu_inf = 0
   for test in range(nb_tests):
      t = vector(ZZ,[D() for i in range(B_L.nrows())])
      v = call_fplll(B_L,t)
      d_2 = RR((v-t).norm(2)/precision)
      d_inf = RR((v-t).norm(infinity)/precision)
      if d_2 > mu_2:
         mu_2 = d_2
      if d_inf > mu_inf:
         mu_inf = d_inf
   return [mu_2, mu_inf]

D = DiscreteGaussianDistributionIntegerSampler(sigma=100*precision)

list_approx_mu = approx_covering_radius([(B_L,20//nb_threads+1,D, 17+i) for i in range(nb_threads)])

approx_mu_2 = 0
approx_mu_inf = 0

for tmp in list_approx_mu:
   approx_mu_2 = max(approx_mu_2,tmp[1][0])
   approx_mu_inf = max(approx_mu_inf,tmp[1][1])


RRout = RealField(13)
print(m, "& ", B_L.nrows(), "& ", RRout(approx_mu_2), "& ",  RRout(approx_mu_inf), "\\\\ ");
print("\\hline");
